/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.jackrabbit.core.persistence.bundle.util;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

/**
 * Same as {@link DbNameIndex} but does not make use of the
 * {@link Statement#RETURN_GENERATED_KEYS} feature as it might not be provided
 * by the underlying database (e.g. oracle).
 */
public class NGKDbNameIndex extends DbNameIndex {

    /**
     * The CVS/SVN id
     */
    static final String CVS_ID = "$URL: http://svn.apache.org/repos/asf/jackrabbit/tags/1.3.1/jackrabbit-core/src/main/java/org/apache/jackrabbit/core/persistence/bundle/util/NGKDbNameIndex.java $ $Rev: 517626 $ $Date: 2007-03-13 12:00:04 +0200 (Tue, 13 Mar 2007) $";

    /**
     * Creates a new index that is stored in a db.
     * @param con the jdbc connection
     * @param schemaObjectPrefix the prefix for table names
     * @throws SQLException if the statements cannot be prepared.
     */
    public NGKDbNameIndex(Connection con, String schemaObjectPrefix)
            throws SQLException {
        super(con, schemaObjectPrefix);
    }

    /**
     * {@inheritDoc}
     */
    protected void init(Connection con, String schemaObjectPrefix)
            throws SQLException {
        nameSelect = con.prepareStatement("select NAME from " + schemaObjectPrefix + "NAMES where ID = ?");
        indexSelect = con.prepareStatement("select ID from " + schemaObjectPrefix + "NAMES where NAME = ?");
        nameInsert = con.prepareStatement("insert into " + schemaObjectPrefix + "NAMES (NAME) values (?)");
    }

    /**
     * Inserts a string into the database and returns the new index.
     * <p/>
     * Instead of using the {@link Statement#RETURN_GENERATED_KEYS} feature, the
     * newly inserted index is retrieved by a 2nd select statement.
     *
     * @param string the string to insert
     * @return the new index.
     */
    protected int insertString(String string) {
        // assert index does not exist
        PreparedStatement stmt = nameInsert;
        ResultSet rs = null;
        try {
            stmt.setString(1, string);
            stmt.executeUpdate();
            return getIndex(string);
        } catch (Exception e) {
            throw new IllegalStateException("Unable to insert index: " + e);
        } finally {
            closeResultSet(rs);
            resetStatement(stmt);
        }
    }
}